"""Documentation utilities."""


def inherit_docs(cls):
    """Class decorator that makes it inherit function doc strings."""
    for name in dir(cls):
        member = getattr(cls, name)
        if member.__doc__ is not None:
            continue

        for parent in cls.mro()[1:]:
            if hasattr(parent, name) and getattr(parent, name).__doc__:
                try:
                    member.__doc__ = getattr(parent, name).__doc__
                except AttributeError:
                    pass

    return cls



"""Construction of parameter spaces."""

import collections
import itertools

from six import string_types

from psyrun.utils.doc import inherit_docs


def dict_concat(args):
    """Concatenates elements with the same key in the passed dictionaries.

    Parameters
    ----------
    args : sequenece of dict
        Dictionaries with sequences to concatenate.

    Returns
    -------
    dict
        The dictionary with the union of all the keys of the dictionaries
        passed in and elements with the same key concatenated. Missing elements
        will be None.

    Examples
    --------
    >>> from pprint import pprint
    >>> pprint(dict_concat(({'a': 0, 'b': 0}, {'a': 1})))
    {'a': [0, 1], 'b': [0, None]}
    """
    keys = set()
    for a in args:
        keys = keys.union(a.keys())
    return {k: [a.get(k, None) for a in args] for k in keys}


class ParameterSpace(collections.Sized):
    """Abstract base class for objects representing a parameter space.

    Supports addition, subtraction, and multiplication operators to construct
    more complicated parameter spaces.

    Deriving classes are supposed to implement the `iterate` and ``__len__``
    methods.


    Parameters
    ----------
    keys : sequence of strings
        Parameter names.
    """

    def __init__(self, keys):
        self._keys = list(keys)

    def build(self):
        """Builds the parameter space into a dictionary of parameter lists.

        Returns
        -------
        dict
            A dictionary with the parameter names as keys and lists with the
            parameter values.

        Examples
        --------
        >>> from pprint import pprint
        >>> pprint(Param(a=[1, 2], b=[1, 2]).build())
        {'a': [1, 2], 'b': [1, 2]}
        """
        built = dict_concat(list(self.iterate()))
        for k in self.keys():
            built.setdefault(k, [])
        return built

    def iterate(self):
        """Iterates over the parameter assignments in the parameter space."""
        raise NotImplementedError()

    def keys(self):
        """Returns the parameter names."""
        return self._keys

    def __len__(self):
        raise NotImplementedError()

    def __add__(self, other):
        return Sum(self, other)

    def __mul__(self, other):
        return Product(self, other)

    def __sub__(self, other):
        return Difference(self, other)

    def __repr__(self):
        keys = sorted(self.keys())
        built = self.build()
        return "Param(**{{{params}}})".format(
            params=", ".join(
                "{k!r}: {v!r}".format(k=k, v=built[k]) for k in keys))

    def __str__(self):
        keys = sorted(self.keys())
        built = self.build()
        return "Param({params})".format(
            params=", ".join(
                "{k!s}={v!r}".format(k=k, v=built[k]) for k in keys))


@inherit_docs
class Param(ParameterSpace):
    """Constructs a simple parameter space from constructor arguments.

    Supports addition, subtraction, and multiplication operators to construct
    more complicated parameter spaces.

    Parameters
    ----------
    params :
        Each keyword argument defines a parameter with a sequence of parameter
        values for it. The length of all lists has to be equal. If a scalar
        instead of a sequence is passed in, it will be replicated to match the
        length of the other parameters. At least one keyword argument has to be
        a sequence.
    """
    def __init__(self, **params):
        super(Param, self).__init__(params.keys())
        self._params = params

        # Make sure strings etc are in a list.
        for k in self._params:
            if isinstance(self._params[k], string_types + (bytes,)):
                self._params[k] = [self._params[k]]

        self._len = None
        for v in self._params.values():
            try:
                l = len(v)
            except TypeError:
                pass
            else:
                if self._len is None:
                    self._len = l
                elif self._len != l:
                    raise ValueError("Parameter lists differ in length.")
        if self._len is None:
            self._len = 1 if len(self._params) > 0 else 0

    def iterate(self):
        for i in range(len(self)):
            yield {k: self.get_param(k, i) for k in self.keys()}

    def __len__(self):
        return self._len

    def get_param(self, key, i):
        """Return the *i*-th parameter assignment.

        Parameters
        ----------
        key : str
            Parameter name of parameter to retrieve.
        i : int
            Index of assigned value to return.
        """
        p = self._params[key]
        try:
            return p[i]
        except TypeError:
            return p


@inherit_docs
class Difference(ParameterSpace):
    """Implements the difference of two parameter spaces.

    Parameters
    ----------
    minuend : `ParameterSpace`
        Minuend (left operand).
    subtrahend : `ParameterSpace`
        Subtrahend (right operand).

    Attributes
    ----------
    minuend : `ParameterSpace`
        Minuend (left operand).
    subtrahend : `ParameterSpace`
        Subtrahend (right operand).

    Examples
    --------
    >>> from pprint import pprint
    >>> pprint(Difference(Param(a=[1, 2], b=[1, 2]), Param(a=[1])).build())
    {'a': [2], 'b': [2]}
    """

    def __init__(self, minuend, subtrahend):
        super(Difference, self).__init__(minuend.keys())
        for k in subtrahend.keys():
            if k not in self._keys:
                raise AmbiguousOperationError(
                    'Key `{0}` not existent in minuend.'.format(k))
        self.left = minuend
        self.right = subtrahend
        self._cached = None

    def iterate(self):
        if len(self.right) == 0:
            return self.left.iterate()
        if self._cached is None:
            exclude = self.right.build()
            self._cached = [item for item in self.left.iterate()
                            if not all(item[k] in exclude[k]
                                       for k in exclude.keys())]
        return iter(self._cached)

    def __len__(self):
        return sum(1 for item in self.iterate())


@inherit_docs
class Product(ParameterSpace):
    """Implements the Cartesian product of two parameter spaces.

    Parameters
    ----------
    left : `ParameterSpace`
        Left operand.
    right : `ParameterSpace`
        Right operand.

    Attributes
    ----------
    left : `ParameterSpace`
        Left operand.
    right : `ParameterSpace`
        Right operand.

    Examples
    --------
    >>> from pprint import pprint
    >>> pprint(Product(Param(a=[1, 2]), Param(b=[1, 2])).build())
    {'a': [1, 1, 2, 2], 'b': [1, 2, 1, 2]}
    """
    def __init__(self, left, right):
        shared_keys = set(left.keys()).intersection(set(right.keys()))
        if len(shared_keys) > 0:
            raise AmbiguousOperationError(
                'Duplicate param keys: {0}'.format(shared_keys))

        super(Product, self).__init__(list(left.keys()) + list(right.keys()))
        self.left = left
        self.right = right

    def iterate(self):
        if len(self.left.keys()) == 0:
            return self.right.iterate()
        elif len(self.right.keys()) == 0:
            return self.left.iterate()
        else:
            return (self._merge(*item) for item in itertools.product(
                self.left.iterate(), self.right.iterate()))

    @staticmethod
    def _merge(left, right):
        merged = {}
        merged.update(left)
        merged.update(right)
        return merged

    def __len__(self):
        if len(self.left.keys()) == 0:
            return len(self.right)
        elif len(self.right.keys()) == 0:
            return len(self.left)
        else:
            return len(self.left) * len(self.right)


@inherit_docs
class Sum(ParameterSpace):
    """Implements the concatenation of two parameter spaces.

    Parameters
    ----------
    left : `ParameterSpace`
        Left operand.
    right : `ParameterSpace`
        Right operand.

    Attributes
    ----------
    left : `ParameterSpace`
        Left operand.
    right : `ParameterSpace`
        Right operand.

    Examples
    --------
    >>> from pprint import pprint
    >>> pprint(Sum(Param(a=[1]), Param(a=[2])).build())
    {'a': [1, 2]}
    """
    def __init__(self, left, right):
        super(Sum, self).__init__(set(left.keys()).union(set(right.keys())))
        self.left = left
        self.right = right

    def iterate(self):
        return ({k: item.get(k, float('nan')) for k in self.keys()}
                for item in itertools.chain(
                    self.left.iterate(), self.right.iterate()))

    def __len__(self):
        return len(self.left) + len(self.right)


def missing(minuend, subtrahend):
    """Return set of parameter assignments missing from another set.

    This differs from a simple subtraction by allowing additional keys in the
    subtrahend, but no additional keys in the minuend.

    Parameters
    ----------
    minuend : `ParameterSpace`
        Parameter space with all assignments.
    subtrahend : :class:`Param`
        Parameter space with assignments to remove from the parameter space.

    Returns
    -------
    `ParameterSpace`
        The reduced parameter space.

    Examples
    --------
    >>> from pprint import pprint
    >>> pprint(missing(Param(a=[1, 2, 3]), Param(a=[2])).build())
    {'a': [1, 3]}
    """
    if len(subtrahend) <= 0:
        return minuend
    for k in minuend.keys():
        if k not in subtrahend.keys():
            raise AmbiguousOperationError()
    return minuend - Param(
        **{k: v for k, v in subtrahend.build().items() if k in minuend.keys()})


class AmbiguousOperationError(RuntimeError):
    """Attempt to combine two parameter spaces in an ambiguous way."""
    pass



"""Management of virtual environments."""

import os


def init_virtualenv(venv=None):
    """Activates a virtual environment.

    This function won't initialize a virtual environment when run on Travis-CI
    which is detected with the ``CI`` environment variable.

    Parameters
    ----------
    venv : str, optional
        Name of virtualenv to activate. If None, the virtualenv given in
        the environment variable ``VIRTUAL_ENV`` will be used. If that
        environment variable is not set either, no virtualenv will be
        activated.
    """
    if os.environ.get('CI', False):
        return

    if venv is None:
        venv = os.environ.get('VIRTUAL_ENV', None)

    if venv is not None:
        activate_script = os.path.join(venv, 'bin', 'activate_this.py')
        if not os.path.exists(activate_script):
            activate_script = os.path.join(venv, 'Scripts', 'activate_this.py')

        with open(activate_script, 'r') as f:
            source = f.read()
        exec(
            compile(source, activate_script, 'exec'),
            dict(__file__=activate_script))



"""NumPy shape handling utility functions."""

import numpy as np


def min_shape(args):
    """Returns the minimum shape that encompasses the shape of all *args*."""
    return tuple(max(x) for x in zip(*args))


def match_shape(a, shape):
    """Make *a* match the shape *shape*."""

    a = np.asarray(a)
    if a.shape == ():
        a = np.asarray([a])
    if a.shape[1:] == shape:
        return a

    dtype = a.dtype
    matched = np.empty((a.shape[0],) + shape, dtype=dtype)
    if np.issubdtype(dtype, float) or np.issubdtype(dtype, complex):
        matched.fill(np.nan)
    elif dtype.kind in 'SU':
        matched.fill('')
    elif dtype.kind == 'O':
        matched.fill(None)
    else:
        raise ValueError("Dtype does not support missing values.")

    matched[np.ix_(*(range(x) for x in a.shape))] = a
    return matched



"""Psyrun's task loading API."""

from __future__ import print_function

import os
import os.path
import re
import sys
import traceback
import warnings

from psyrun.backend import DefaultBackend
from psyrun.pspace import Param
from psyrun.store import DefaultStore
from psyrun.scheduler import ImmediateRun


class TaskDef(object):
    """Task defined by a Python file.

    Parameters
    ----------
    path : str
        Python file to load as task.
    conf : `Config`
        Default values for task parameters.

    Attributes
    ----------
    TASK_PATTERN : re.RegexObject
        Regular expression to match task filenames.
    """

    TASK_PATTERN = re.compile(r'^task_(.*)$')

    def __init__(self, path, conf=None):
        if conf is None:
            taskdir = os.path.dirname(path)
            conffile = os.path.join(taskdir, 'psy-conf.py')
            if os.path.exists(conffile):
                conf = Config.load_from_file(conffile)
            else:
                conf = Config()

        _set_public_attrs_from_dict(
            self, _load_pyfile(path), only_existing=False)
        self.path = path

        if not hasattr(self, 'name'):
            prefixed_name, _ = os.path.splitext(os.path.basename(path))
            m = self.TASK_PATTERN.match(prefixed_name)
            if m:
                self.name = m.group(1)
            else:
                self.name = prefixed_name

        conf.apply_as_default(self)


def _load_pyfile(filename):
    source = ''
    with open(filename, 'r') as f:
        source += f.read()
    code = compile(source, filename, 'exec')
    loaded = {'__file__': filename}
    exec(code, loaded)  # pylint: disable=exec-used
    return loaded


def _set_public_attrs_from_dict(obj, d, only_existing=True):
    for k, v in d.items():
        if not k.startswith('_') and (not only_existing or hasattr(obj, k)):
            setattr(obj, k, v)


class Config(object):  # pylint: disable=too-many-instance-attributes
    """Task configuration.


    Attributes
    ----------
    backend : `Backend`, default: `DistributeBackend`
        The processing backend which determines how work is distributed across
        jobs.
    file_dep : sequence of str, default: ``[]``
        Additional files the task depends on.
    max_jobs : int, default: 100
        Maximum number of jobs to start. With less jobs each job has to process
        more parameter assignments. It depends on the  scheduler and backend
        used to which degree these will run in parallel.
    min_items : int, default: 1
        Minimum number of parameter assignment to evaluate per job. If a single
        assignment is fast to evaluate, increasing this number can improve
        performance because Psyrun will not start a new job for each parameter
        assignment which can save some overhead.
    overwrite_dirty : bool, default: True
        Whether to overwrite dirty workdirs without a warning.
    pspace : `ParameterSpace`, required
        Parameter space to evaluate.
    python : str, default: ``sys.executable``
        Path to Python interpreter to use.
    resultfile : str or None, default: None
        Path to save the results of the finished task at. If None, this
        defaults to ``'result.<ext>'`` in the *workdir*.
    scheduler : `Scheduler`, default: `ImmediateRun`
        Scheduler to use to submit individual jobs.
    scheduler_args : dict, default: ``{}``
        Additional scheduler arguments. See the documentation of the
        scheduler for details.
    store : `Store`, default: `PickleStore`
        Input/output backend.
    workdir : str, default: ``'psy-work'``
        Working directory to store results and supporting data to process the
        task.

    """

    __slots__ = [
        'backend', 'file_dep', 'max_jobs', 'min_items', 'pspace',
        'overwrite_dirty', 'python', 'resultfile', 'scheduler',
        'scheduler_args', 'store', 'workdir']

    def __init__(self):
        self.backend = DefaultBackend
        self.file_dep = []
        self.max_jobs = 100
        self.min_items = 1
        self.overwrite_dirty = True
        self.pspace = Param()
        self.python = sys.executable
        self.resultfile = None
        self.scheduler = ImmediateRun()
        self.scheduler_args = dict()
        self.store = DefaultStore()
        self.workdir = os.path.abspath('psy-work')

    @classmethod
    def load_from_file(cls, filename):
        """Load the configuration values from a Python file.

        Parameters
        ----------
        filename : str
            Python file to load.
        """
        conf = cls()
        loaded_conf = _load_pyfile(filename)
        _set_public_attrs_from_dict(conf, loaded_conf)
        return conf

    def apply_as_default(self, task):
        """Copies the attributes to a different object given they are not set
        in that object.

        Parameters
        ----------
        task : obj
            Object to copy the attributes to.
        """
        for attr in self.__slots__:
            if not hasattr(task, attr):
                setattr(task, attr, getattr(self, attr))


class PackageLoader(object):
    """Loads tasks from Python files.

    Filenames have to match the regular expression defined in
    `TaskDef.TASK_PATTERN`. See `Config` for supported module
    level variables in the task definition.

    It is possible to set these variables for all tasks by setting them in
    the file ``psy-conf.py`` in the *taskdir*.

    Parameters
    ----------
    taskdir : str
        Directory to load task files from.

    Attributes
    ----------
    taskdir : str
        Directory to load task files from.
    conf : `Config`
        Default values for module level task variables.
    """
    def __init__(self, taskdir):
        super(PackageLoader, self).__init__()
        self.taskdir = taskdir
        conffile = os.path.join(self.taskdir, 'psy-conf.py')
        if os.path.exists(conffile):
            self.conf = Config.load_from_file(conffile)
        else:
            self.conf = Config()

    def load_task_defs(self):
        """Load task definitions.

        Returns
        -------
        list of `TaskDef`
            Task definitions.
        """
        task_defs = []
        for filename in os.listdir(self.taskdir):
            root, ext = os.path.splitext(filename)
            if TaskDef.TASK_PATTERN.match(root) and ext == '.py':
                path = os.path.join(self.taskdir, filename)
                try:
                    task_defs.append(TaskDef(path, self.conf))
                except Exception:  # pylint: disable=broad-except
                    traceback.print_exc()
                    warnings.warn("Task {path!r} could not be loaded.".format(
                        path=path))
        return task_defs



"""Store using pickle files."""

import errno

from six import string_types
from six.moves import cPickle as pickle

from psyrun.store.base import Store
from psyrun.utils.doc import inherit_docs


@inherit_docs
class PickleStore(Store):
    """Store using Python pickle *.pkl* files.

    It supports all pickle-able data types and has no additional dependencies,
    but is not the most efficient store. Also, it has to load the complete
    file to append to it.
    """

    ext = '.pkl'

    def __init__(self, protocol=pickle.HIGHEST_PROTOCOL):
        self.protocol = protocol

    def save(self, filename, data):
        with open(filename, 'wb') as f:
            pickle.dump(data, f, self.protocol)

    def load(self, filename, row=None):
        with open(filename, 'rb') as f:
            data = pickle.load(f)
        if row is None:
            return data
        else:
            return {k: [v[row]] for k, v in data.items()}

    def _get_data_len(self, data):
        if len(data) > 0:
            return max(
                len(v) if not isinstance(v, string_types + (bytes,)) else 1
                for v in data.values())
        else:
            return 0

    def append(self, filename, data):
        try:
            loaded = self.load(filename)
        except IOError as err:
            if err.errno != errno.ENOENT:
                raise
            self.save(filename, data)
        else:
            keys = set(loaded.keys())
            keys = keys.union(data.keys())

            old_n = self._get_data_len(loaded)
            new_n = self._get_data_len(data)

            for k in keys:
                if k not in loaded:
                    loaded[k] = [None] * old_n
                if not isinstance(loaded[k], list):
                    loaded[k] = list(loaded[k])

                v = data.get(k, [None])
                if len(v) != new_n:
                    if len(v) == 1:
                        v = v * new_n
                    else:
                        raise ValueError("Incompatible data length.")
                loaded[k].extend(v)

            self.save(filename, loaded)



"""Testing utilities."""

import os.path
import pickle
import subprocess

from psyrun.scheduler import JobStatus, Scheduler
from psyrun.utils.doc import inherit_docs


@inherit_docs
class MockScheduler(Scheduler):
    """Mock scheduler implementation.

    Parameters
    ----------
    datafile : str
        File to store supporting data.
    """

    def __init__(self, datafile):
        self.datafile = datafile

    @property
    def next_id(self):
        """ID for next submitted job."""
        if os.path.exists(self.datafile):
            with open(self.datafile, 'rb') as f:
                return pickle.load(f)['next_id']
        else:
            return 0

    @next_id.setter
    def next_id(self, value):
        self._serialize(next_id=value)

    @property
    def joblist(self):
        """Tuple of current jobs."""
        if os.path.exists(self.datafile):
            with open(self.datafile, 'rb') as f:
                return tuple(pickle.load(f)['joblist'])
        else:
            return tuple()

    @joblist.setter
    def joblist(self, value):
        self._serialize(joblist=value)

    def _serialize(self, next_id=None, joblist=None):
        if next_id is None:
            next_id = self.next_id
        if joblist is None:
            joblist = self.joblist

        with open(self.datafile, 'wb') as f:
            pickle.dump(
                {'next_id': next_id, 'joblist': joblist},
                f, pickle.HIGHEST_PROTOCOL)

    def submit(
            self, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        if depends_on is None:
            depends_on = []
        jobid = self.next_id
        self.next_id += 1
        self.joblist = self.joblist + ({
            'id': jobid,
            'args': args,
            'output_filename': output_filename,
            'name': name or str(jobid),
            'depends_on': depends_on,
            'scheduler_args': scheduler_args,
            'status': '*Q' if len(depends_on) > 0 else 'Q',
        },)
        return jobid

    def mark_running(self):
        """Mark all jobs as running."""
        updated = []
        for job in self.joblist:
            job.update({'status': 'R'})
            updated.append(job)
        self.joblist = tuple(updated)

    def kill(self, jobid):
        self.joblist = [job for job in self.joblist if job['id'] != jobid]

    def get_status(self, jobid):
        for job in self.joblist:
            if job['id'] == jobid:
                return JobStatus(job['id'], job['status'], job['name'])
        return None

    def consume(self):
        """Process all queued jobs."""
        for job in self.joblist:
            with open(job['output_filename'], 'a') as f:
                subprocess.check_call(
                    job['args'], stdout=f, stderr=subprocess.STDOUT)
        self.joblist = []

    def consume_job(self, job):
        """Process a  single job."""
        with open(job['output_filename'], 'a') as f:
            subprocess.check_call(
                job['args'], stdout=f, stderr=subprocess.STDOUT)
        self.joblist = [j for j in self.joblist if j != job]

    def get_jobs(self):
        return [job['id'] for job in self.joblist]



"""Psyrun specific exceptions and warnings."""

import warnings


class TaskWarning(UserWarning):
    """General warning related to task processing.

    *TaskWarnings* will always be shown by default.
    """
    pass


class JobsRunningWarning(TaskWarning):
    """Warning issued when jobs for a task are still running."""
    def __init__(self, name):
        self.task = name
        super(JobsRunningWarning, self).__init__(
            "Task '{}' has unfinished jobs queued. Not starting new jobs "
            "until these are finished or have been killed.".format(name))


class TaskWorkdirDirtyWarning(TaskWarning):
    """Warning issued when the workdir is dirty and would be overwritten."""
    def __init__(self, name):
        # TODO explain how to solve this
        self.task = name
        super(TaskWorkdirDirtyWarning, self).__init__(
            "Work directory of task '{}' is dirty.".format(name))


warnings.simplefilter('always', category=TaskWarning)



from psyrun.backend import DistributeBackend, LoadBalancingBackend
from psyrun.mapper import (
    map_pspace,
    map_pspace_parallel,
    map_pspace_hdd_backed)
from psyrun.pspace import Param
from psyrun.scheduler import ImmediateRun, Sqsub
from psyrun.store import DefaultStore, PickleStore
from psyrun.version import version as __version__



"""Defines the ``psy`` command line interface commands."""

import argparse
import os.path
import shutil
import sys
import warnings

from psyrun.backend.distribute import Splitter
from psyrun.exceptions import TaskWorkdirDirtyWarning
from psyrun.jobs import Clean, Fullname, Submit, Uptodate
from psyrun.tasks import PackageLoader
from psyrun.utils.venv import init_virtualenv


commands = {}


def psy_main(argv=None, init_venv=True):
    """Runs psyrun tasks.

    Parameters
    ----------
    argv : sequence of str, optional
        ``psy`` command line arguments.
    init_venv : bool, optional
        Use the virtualenv active in the shell environment if set to True.

    Returns
    -------
    int
        Return code.
    """
    if init_venv:
        init_virtualenv()

    parser = argparse.ArgumentParser(
        epilog="available commands:\n{}\n".format(''.join(
            "  {: <11s} {}\n".format(k, v.short_desc)
            for k, v in sorted(commands.items()))),
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('cmd', nargs=1, type=str, help="command to run")
    parser.add_argument('args', nargs=argparse.REMAINDER,
                        help="arguments to command")
    args = parser.parse_args(argv)

    return commands[args.cmd[0]](args.cmd[0], args.args).run()


class Command(object):
    """Base class for commands.

    Deriving classes are supposed to implement `run` and may overwrite
    `add_args` to add additional arguments to ``self.parser``.

    Parameters
    ----------
    cmd : str
        Command name.
    argv : sequence
        Arguments to the command.

    Attributes
    ----------
    parser : argparse.ArgumentParser
        Parser for arguments.
    args : argparse.Namespace
        Parsed arguments.
    """
    short_desc = ""
    long_desc = ""

    def __init__(self, cmd, argv):
        prog = os.path.basename(sys.argv[0]) + ' ' + cmd
        self.parser = argparse.ArgumentParser(
            prog=prog, description=self.long_desc)
        self.add_args()
        self.args = self.parser.parse_args(argv)

    def add_args(self):
        """Add command arguments to ``self.parser``."""
        pass

    def run(self):
        """Run the command."""
        raise NotImplementedError()


class TaskdirCmd(Command):
    """Base class for commands that accept a ``--taskdir`` argument."""

    def __init__(self, cmd, argv):
        super(TaskdirCmd, self).__init__(cmd, argv)
        self.package_loader = PackageLoader(self.args.taskdir[0])

    def add_args(self):
        super(TaskdirCmd, self).add_args()
        self.parser.add_argument(
            '--taskdir', nargs=1, type=str, default=['psy-tasks'],
            help="directory to load tasks from")

    def run(self):
        raise NotImplementedError()


class TaskselCmd(TaskdirCmd):
    """Base class for commands that accept a selection of tasks as arguments.

    Attributes
    ----------
    default_to_all : bool
        Indicates whether the command defaults to all (or no) tasks when no
        task names are specified as arguments.
    """

    default_to_all = True

    def add_args(self):
        super(TaskselCmd, self).add_args()
        self.parser.add_argument('task', nargs='*', type=str)

    def run(self):
        tasks = {t.name: t for t in self.package_loader.load_task_defs()}
        run_all = self.default_to_all and len(self.args.task) == 0
        if run_all:
            selected = tasks.keys()
        else:
            selected = self.args.task
        for name in selected:
            self.run_task(tasks[name])

    def run_task(self, task):
        raise NotImplementedError()


class RunCmd(TaskselCmd):
    short_desc = "run one or more tasks"
    long_desc = (
        "Submits the jobs to run one or more tasks. If no task name is "
        "provided, all tasks will be run.")

    def add_args(self):
        super(RunCmd, self).add_args()
        self.parser.add_argument(
            '-c', '--continue', action='store_true',
            help="preserve existing data in the workdir and do not rerun "
            "parameter assignment already existent in there")

    def run_task(self, task):
        cont = vars(self.args)['continue']
        backend = task.backend(task)

        job = backend.create_job(cont=cont)
        names = Fullname(job).names
        uptodate = Uptodate(job, names, task)

        if not cont:
            if uptodate.outdated and os.path.exists(backend.resultfile):
                if not task.overwrite_dirty:
                    warnings.warn(TaskWorkdirDirtyWarning(task.name))
                    return

            Clean(job, task, names, uptodate)
        Submit(job, names, uptodate)


class CleanCmd(TaskselCmd):
    short_desc = "clean task data"
    long_desc = (
        "Removes all processing and result files associated with a task.")
    default_to_all = False

    def run_task(self, task):
        path = os.path.join(task.workdir, task.name)
        print('rm', path)
        shutil.rmtree(path)


class ListCmd(TaskdirCmd):
    short_desc = "list tasks"
    long_desc = "Lists all available tasks."

    def run(self):
        for t in self.package_loader.load_task_defs():
            print(t.name)


class MergeCmd(Command):
    short_desc = "merge data files into single file"
    long_desc = "Merges all data files in a directory into a single data file."

    def add_args(self):
        self.parser.add_argument(
            'directory', type=str, help="directory with files to merge")
        self.parser.add_argument(
            'merged', type=str, help="file to write the merged result to")

    def run(self):
        ext = os.path.splitext(self.args.merged)[1].lower()
        if ext == '.npz':
            from psyrun.store.npz import NpzStore as Store
        elif ext == '.h5':
            from psyrun.store.h5 import H5Store as Store
        else:
            from psyrun.store.pickle import PickleStore as Store
        Splitter.merge(self.args.directory, self.args.merged, store=Store())


class StatusCmd(TaskselCmd):
    short_desc = "print status of tasks"
    long_desc = (
        "Prints the status (number of completed parameter assignments of the "
        "total). Can also print parameter assignments that have not been "
        "evaluated yet.")

    def add_args(self):
        super(StatusCmd, self).add_args()
        self.parser.add_argument(
            '-v', '--verbose', action='store_true',
            help="print missing parameter sets")

    def run_task(self, task):
        missing = task.backend(task).get_missing()
        print("{}:".format(task.name))
        print("  {} out of {} rows completed.".format(
            len(task.pspace) - len(missing), len(task.pspace)))
        if self.args.verbose:
            print("  Missing parameter sets:")
            for pset in missing.iterate():
                print('   ', pset)
        print("")


class TestCmd(TaskselCmd):
    short_desc = "test execution of single parameter assignment"
    long_decs = (
        "Tests the task execution by running a single parameter assignment. "
        "The job will not be submitted with to the scheduler, but run "
        "immediatly.")

    def run_task(self, task):
        print(task.name)
        task.execute(**next(task.pspace.iterate()))


commands.update({
    'run': RunCmd,
    'clean': CleanCmd,
    'list': ListCmd,
    'merge': MergeCmd,
    'status': StatusCmd,
    'test': TestCmd,
})


if __name__ == '__main__':
    sys.exit(psy_main())



"""Handling and processing of job trees."""

import itertools
import os
import os.path
import warnings

from psyrun.exceptions import JobsRunningWarning
from psyrun.utils.doc import inherit_docs


class Job(object):
    """Describes a single processing job.

    Parameters
    ----------
    name : str
        Name of the job.
    submit_fn : function
        Function to use to submit the job for processing.
    code : str
        Python code to execute.
    dependencies : sequence
        Identifiers of other jobs that need to finish first before this job
        can be run.
    targets : sequence of str
        Files created by this job.

    Attributes
    ----------
    name : str
        Name of the job.
    submit_fn : function
        Function to use to submit the job for processing.
    code : str
        Python code to execute.
    dependencies : sequence
        Identifiers of other jobs that need to finish first before this job
        can be run.
    targets : sequence of str
        Files created by this job.
    """
    def __init__(self, name, submit_fn, code, dependencies, targets):
        self.name = name
        self.submit_fn = submit_fn
        self.code = code
        self.dependencies = dependencies
        self.targets = targets


class JobChain(object):
    """Chain of jobs to run in succession.

    Parameters
    ----------
    name : str
        Name of the job chain.
    jobs : sequence of Job
        Jobs to run in succession.

    Attributes
    ----------
    name : str
        Name of the job chain.
    jobs : sequence of Job
        Jobs to run in succession.
    dependencies : sequence
        Jobs that need to run first before the job chain can be run (equivalent
        to the dependencies of the first job in the chain).
    targets : sequence of str
        Files created or updated by the job chain (equivalent to the targets
        of the last job in the chain).
    """
    def __init__(self, name, jobs):
        self.name = name
        self.jobs = jobs

    @property
    def dependencies(self):
        return self.jobs[0].dependencies

    @property
    def targets(self):
        return self.jobs[-1].targets


class JobGroup(object):
    """Group of jobs that can run in parallel.

    Parameters
    ----------
    name : str
        Name of the job group.
    jobs : sequence of Job
        Jobs to run in the job group.

    Attributes
    ----------
    name : str
        Name of the job group.
    jobs : sequence of Job
        Jobs to run in the job group.
    dependencies : sequence
        Jobs that need to run first before the job group can be run (equivalent
        to the union of all the group's job's dependencies).
    targets : sequence of str
        Files that will be created or updated by the group's jobs (equivalent
        to the union of all the group's job's targets).
    """
    def __init__(self, name, jobs):
        self.name = name
        self.jobs = jobs

    @property
    def dependencies(self):
        return itertools.chain(j.dependencies for j in self.jobs)

    @property
    def targets(self):
        return itertools.chain.from_iterable(j.targets for j in self.jobs)


class JobTreeVisitor(object):
    """Abstract base class to implement visitors on trees of jobs.

    Base class to implement visitors following the Visitor pattern to traverse
    the tree constructed out of `Job`, `JobChain`, and `JobGroup` instances.

    A deriving class should overwrite `visit_job`, `visit_chain`, and
    `visit_group`. Use the `visit` method to start visiting a tree of jobs.
    """

    def __init__(self):
        self._dispatcher = {
            Job: self.visit_job,
            JobChain: self.visit_chain,
            JobGroup: self.visit_group
        }

    def visit_job(self, job):
        raise NotImplementedError()

    def visit_chain(self, chain):
        raise NotImplementedError()

    def visit_group(self, group):
        raise NotImplementedError()

    def visit(self, job):
        """Visit all jobs in the tree *job*."""
        return self._dispatcher[job.__class__](job)


@inherit_docs
class Submit(JobTreeVisitor):
    """Submit all jobs that are not up-to-date.

    The constructor will call `visit`.

    Parameters
    ----------
    job : job tree
        Tree of jobs to submit.
    names : dict
        Maps jobs to their names. (Can be obtained with `Fullname`.)
    uptodate : dict
        Maps jobs to their up-to-date status.
        (Can be obtained with `Uptodate`.)

    Attributes
    ----------
    names : dict
        Maps jobs to their names.
    uptodate : dict
        Maps jobs to their up-to-date status.
    """
    def __init__(self, job, names, uptodate):
        super(Submit, self).__init__()
        self.names = names
        self.uptodate = uptodate
        self._depends_on = []
        self.visit(job)

    def visit_job(self, job):
        if self.uptodate.status[job]:
            print('-', self.names[job])
            return []
        else:
            print('.', self.names[job])
            return [job.submit_fn(
                job.code, self.names[job], depends_on=self._depends_on)]

    def visit_group(self, group):
        return sum((self.visit(job) for job in group.jobs), [])

    def visit_chain(self, chain):
        old_depends_on = self._depends_on
        job_ids = []
        for job in chain.jobs:
            ids = self.visit(job)
            job_ids.extend(ids)
            self._depends_on = old_depends_on + ids
        self._depends_on = old_depends_on
        return job_ids


@inherit_docs
class Clean(JobTreeVisitor):
    """Clean all target files and supporting files of jobs that are outdated.

    The constructor will call visit.

    Parameters
    ----------
    job : job tree
        Tree of jobs to clean.
    task : TaskDef
        Task that generated the job tree.
    names : dict
        Maps jobs to their names. (Can be obtained with Fullname.)
    uptodate : dict, optional
        Maps jobs to their up-to-date status.
        (Can be obtained with Uptodate.)
        If not provided, all jobs are treated as outdated.

    Attributes
    ----------
    task : TaskDef
        Task that generated the job tree.
    names : dict
        Maps jobs to their names.
    uptodate : dict
        Maps jobs to their up-to-date status.
    """
    def __init__(self, job, task, names, uptodate=None):
        super(Clean, self).__init__()
        self.task = task
        self.names = names
        if uptodate is None:
            uptodate = {}
        self.uptodate = uptodate
        self.visit(job)

    def visit_job(self, job):
        if self.uptodate.get(job, False):
            return

        workdir = os.path.join(self.task.workdir, self.task.name)
        for item in os.listdir(workdir):
            if item.startswith(self.names[job]):
                os.remove(os.path.join(workdir, item))
        for t in job.targets:
            if os.path.exists(t):
                os.remove(t)

    def visit_chain(self, chain):
        pass

    def visit_group(self, chain):
        pass


@inherit_docs
class Fullname(JobTreeVisitor):
    """Construct names of the jobs.

    The constructor will call `visit`.

    Parameters
    ----------
    jobtree : job tree
        Tree of jobs to construct names for.

    Attributes
    ----------
    names : dict
        Maps jobs to their names.
    """
    def __init__(self, jobtree):
        super(Fullname, self).__init__()
        self._prefix = ''
        self.names = {}
        self.visit(jobtree)

    def visit_job(self, job):
        self.names[job] = self._prefix + job.name

    def visit_chain(self, chain):
        self.visit_group(chain)

    def visit_group(self, group):
        self.names[group] = self._prefix + group.name
        old_prefix = self._prefix
        self._prefix += group.name + ':'
        for job in group.jobs:
            self.visit(job)
        self._prefix = old_prefix


@inherit_docs
class Uptodate(JobTreeVisitor):
    """Determines the up-to-date status of jobs.

    The constructor will call visit.

    Parameters
    ----------
    jobtree : job tree
        Tree of jobs to determine the up-to-date status for.
    names : dict
        Maps jobs to their names. (Can be obtained with Fullname.)
    task : TaskDef
        Task that generated the job tree.

    Attributes
    ----------
    names : dict
        Maps jobs to their names.
    task : TaskDef
        Task that generated the job tree.
    status : dict
        Maps jobs to their up-to-date status.
    """
    def __init__(self, jobtree, names, task):
        super(Uptodate, self).__init__()
        self.names = names
        self.task = task
        self.status = {}
        self._clamp = None

        self.any_queued = False
        self.outdated = False

        self.visit(jobtree)
        self.post_visit()

    def post_visit(self):
        """Called after `visit`.

        Checks whether jobs are still running and marks these as up-to-date
        while issuing a warning.
        """
        skip = False
        if self.any_queued and self.outdated:
            skip = True
            warnings.warn(JobsRunningWarning(self.task.name))
        if skip:
            for k in self.status:
                self.status[k] = True

    def visit_job(self, job):
        if self.is_job_queued(job):
            self.status[job] = True
        elif self._clamp is None:
            tref = self._get_tref(job.dependencies)
            self.status[job] = self.files_uptodate(tref, job.targets)
        else:
            self.status[job] = self._clamp
        return self.status[job]

    def visit_chain(self, chain):
        if self._clamp is None:
            tref = self._get_tref(chain.jobs[0].dependencies)

            last_uptodate = -1
            for i, job in enumerate(reversed(chain.jobs)):
                if self.files_uptodate(tref, job.targets):
                    last_uptodate = len(chain.jobs) - i - 1
                    break

            for i, job in enumerate(chain.jobs):
                if i <= last_uptodate:
                    self._clamp = True
                elif i == last_uptodate + 1:
                    self._clamp = None
                else:
                    self._clamp = False
                self.visit(job)

            self.status[chain] = last_uptodate + 1 == len(chain.jobs)
            self._clamp = None
        else:
            for job in chain.jobs:
                self.visit(job)
            self.status[chain] = self._clamp
        return self.status[chain]

    def visit_group(self, group):
        subtask_status = [self.visit(j) for j in group.jobs]
        self.status[group] = all(subtask_status)
        return self.status[group]

    def is_job_queued(self, job):
        """Checks whether *job* is queud."""
        job_names = [
            self.task.scheduler.get_status(j).name
            for j in self.task.scheduler.get_jobs()]
        is_queued = self.names[job] in job_names
        self.any_queued |= is_queued
        return is_queued

    def files_uptodate(self, tref, targets):
        """Checks that all *targets* are newer than *tref*."""
        uptodate = all(
            self._is_newer_than_tref(target, tref) for target in targets)
        self.outdated |= not uptodate
        return uptodate

    def _get_tref(self, dependencies):
        tref = 0
        deps = [d for d in dependencies if os.path.exists(d)]
        if len(deps) > 0:
            tref = max(os.stat(d).st_mtime for d in deps)
        return tref

    def _is_newer_than_tref(self, filename, tref):
        return os.path.exists(filename) and os.stat(filename).st_mtime >= tref



"""Job scheduler."""

from collections import namedtuple
import os
import os.path
import subprocess


JobStatus = namedtuple('JobStatus', ['id', 'status', 'name'])


class Scheduler(object):
    """Scheduler interface."""

    def submit(
            self, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        """Submit a job.

        Parameters
        ----------
        args : sequence
            The command and arguments to execute.
        output_filename : str
            File to write process output to.
        name : str, optional
            Name of job.
        depends_on : sequence of int, optional
            IDs of jobs that need to finish first before the submitted job can
            be started.
        scheduler_args : dict, optional
            Additional arguments for the scheduler.

        Returns
        -------
        Job ID
        """
        raise NotImplementedError()

    def kill(self, jobid):
        """Kill a job.

        Parameters
        ----------
        jobid
            Job to kill.
        """
        raise NotImplementedError()

    def get_status(self, jobid):
        """Get the status of a job.

        Parameters
        ----------
        jobid
            Job to request status of.

        Returns
        -------
        `JobStatus`
            Returns a tuple with *(id, status, name)* wherein status can be

            * ``'Q'`` for a queued job
            * ``'*Q'`` for a queued job waiting on another job to finish
            * ``'Z'`` for a sleeping job
            * ``'D'`` for a completed job

            If no status data is available for the job ID, None will be
            returned.
        """
        raise NotImplementedError()

    def get_jobs(self):
        """Get all queued, running, and recently finished jobs.

        Returns
        -------
        sequence
            Job IDs
        """
        raise NotImplementedError()


class ImmediateRun(Scheduler):
    """Runs jobs immediately on the local machine."""

    _cur_id = 0
    _failed_jobs = []

    def submit(
            self, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        """Submit a job.

        Parameters
        ----------
        args : sequence
            The command and arguments to execute.
        output_filename : str
            File to write process output to.
        name
            Unused.
        depends_on
            Unused
        scheduler_args
            Unused.

        Returns
        -------
        int
            Job ID
        """
        self._cur_id += 1

        if any(x in self._failed_jobs for x in depends_on):
            self._failed_jobs.append(self._cur_id)
            return self._cur_id

        with open(output_filename, 'a') as f:
            if subprocess.call(args, stdout=f, stderr=subprocess.STDOUT) != 0:
                self._failed_jobs.append(self._cur_id)
        return self._cur_id

    def kill(self, jobid):
        """Has no effect."""
        pass

    def get_status(self, jobid):
        """Has no effect."""
        pass

    def get_jobs(self):
        """Returns an empty list."""
        return []


class Sqsub(Scheduler):
    """sqsub (sharcnet) scheduler."""

    class _Option(object):
        def __init__(self, name, conversion=str):
            self.name = name
            self.conversion = conversion

        def build(self, value):
            raise NotImplementedError()

    class _ShortOption(_Option):
        def build(self, value):
            if value is None:
                return []
            return [self.name, self.conversion(value)]

    class _LongOption(_Option):
        def build(self, value):
            if value is None:
                return []
            return [self.name + '=' + self.conversion(value)]

    KNOWN_ARGS = {
        'timelimit': _ShortOption('-r'),
        'output_file': _ShortOption('-o'),
        'n_cpus': _ShortOption('-n'),
        'n_nodes': _ShortOption('-N'),
        'memory': _LongOption('--mpp'),
        'depends_on': _ShortOption(
            '-w', lambda jobids: ','.join(str(x) for x in jobids)),
        'idfile': _LongOption('--idfile'),
        'name': _ShortOption('-j')
    }

    def build_args(self, **kwargs):
        args = []
        for k, v in kwargs.items():
            args.extend(self.KNOWN_ARGS[k].build(v))
        return args

    def __init__(self, workdir=None):
        if workdir is None:
            workdir = '/work/{user}/psyrun'.format(user=os.environ['USER'])
        if not os.path.exists(workdir):
            os.makedirs(workdir)
        self.workdir = workdir
        self.idfile = os.path.join(workdir, 'idfile')
        self._jobs = None

    def submit(
            self, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        """Submit a job.

        Parameters
        ----------
        args : list
            The command and arguments to execute.
        output_filename : str
            File to write process output to.
        name : str, optional
            Name of job.
        depends_on : list of int, optional
            IDs of jobs that need to finish first before the submitted job can
            be started.
        scheduler_args : dict, optional
            Additional arguments for the scheduler.

        Returns
        -------
        int
            Job ID
        """
        if scheduler_args is None:
            scheduler_args = dict()
        else:
            scheduler_args = dict(scheduler_args)

        # Checking whether jobs depending on are completed before submitting
        # the new job is a race condition, but I see no possibility to avoid
        # it.
        if depends_on is not None:
            statii = [self.get_status(x) for x in depends_on]
            depends_on = [x for x, s in zip(depends_on, statii)
                          if s is not None and s.status != 'D']

        scheduler_args.update({
            'idfile': self.idfile,
            'output_file': output_filename,
            'depends_on': depends_on,
            'name': name,
        })
        subprocess.check_call(
            ['sqsub'] + self.build_args(**scheduler_args) + args)
        with open(self.idfile, 'r') as f:
            return int(f.read())

    def kill(self, jobid):
        """Kill a job.

        Parameters
        ----------
        jobid : int
            Job to kill.
        """
        subprocess.check_call(['sqkill', str(jobid)])

    def get_status(self, jobid=None):
        """Get the status of a job.

        Parameters
        ----------
        jobid : int
            Job to request status of.

        Returns
        -------
        namedtuple or None
            Returns a tuple with *(id, status, name)* wherein status can be

            - ``'Q'`` for a queued job
            - ``'*Q'`` for a queued job waiting on another job to finish
            - ``'Z'`` for a sleeping job
            - ``'D'`` for a completed job

            If no status data is available for the job ID, None will be
            returned.
        """
        if self._jobs is None:
            self.refresh_job_info()
        elif jobid not in self._jobs:
            try:
                stdout = subprocess.check_output(
                    ['sqjobs', str(jobid)], stderr=subprocess.STDOUT,
                    universal_newlines=True)
                for line in stdout.split(os.linesep)[2:]:
                    cols = line.split(None, 5)
                    if len(cols) > 3 and int(cols[0]) == jobid:
                        if cols[2] == 'C':
                            cols[2] = 'D'
                        self._jobs[jobid] = JobStatus(jobid, cols[2], cols[-1])
            except subprocess.CalledProcessError as err:
                # 1 if none pending, running, or suspended
                if err.returncode != 1:
                    raise
        return self._jobs.get(jobid, None)

    def get_jobs(self):
        """Get all queued and running jobs.

        Returns
        -------
        list of int
            Job IDs
        """
        if self._jobs is None:
            self.refresh_job_info()
        return self._jobs.keys()

    def refresh_job_info(self):
        self._jobs = {}
        stdout = subprocess.check_output(['sqjobs'], universal_newlines=True)
        for line in stdout.split('\n')[2:]:
            cols = line.split(None, 5)
            if len(cols) > 2 and cols[2] in ['Q', '*Q', 'Z']:
                jobid = int(cols[0])
                self._jobs[jobid] = JobStatus(jobid, cols[2], cols[-1])



"""Load balancing backend."""

import fcntl
import os

from psyrun.backend.base import Backend
from psyrun.jobs import Job, JobChain, JobGroup
from psyrun.mapper import map_pspace
from psyrun.pspace import missing, Param
from psyrun.store import DefaultStore
from psyrun.utils.doc import inherit_docs


@inherit_docs
class LoadBalancingBackend(Backend):
    """Create subtasks for load balanced parameter evaluation.

    This will create *max_jobs* worker jobs that will fetch parameter
    assignments from a queue. Thus, all worker jobs should be busy all of the
    time.

    This backend is useful if individual parameters assignments can vary to a
    large degree in their processing time. It is recommended to use a `Store`
    that supports efficient fetching of a single row from a file.

    Parameters
    ----------
    task : `TaskDef`
        Task definition to create subtasks for.
    """

    @property
    def infile(self):
        """File from which input parameter assignments are fetched."""
        return os.path.join(self.workdir, 'in' + self.task.store.ext)

    @property
    def statusfile(self):
        """File that stores the processing status."""
        return os.path.join(self.workdir, 'status')

    @property
    def partial_resultfile(self):
        """File results are appended to while processing is in progress."""
        root, ext = os.path.splitext(self.resultfile)
        return root + '.part' + ext

    @property
    def resultfile(self):
        """Final result file."""
        if self.task.resultfile:
            return self.task.resultfile
        else:
            return os.path.join(self.workdir, 'result' + self.task.store.ext)

    def create_job(self, cont=False):
        pspace = self.create_pspace_job(cont=cont)
        process = self.create_process_job()
        finalize = self.create_finalize_job()
        return JobChain(self.task.name, [pspace, process, finalize])

    def create_pspace_job(self, cont):
        code = '''
import os.path
from psyrun.pspace import missing, Param
from psyrun.backend.load_balancing import LoadBalancingWorker
pspace = task.pspace
if {cont!r}:
    if os.path.exists({outfile!r}):
        pspace = missing(pspace, Param(**task.store.load({outfile!r})))
    if os.path.exists({part_outfile!r}):
        pspace = missing(pspace, Param(**task.store.load({part_outfile!r})))
task.store.save({infile!r}, pspace.build())
LoadBalancingWorker.create_statusfile({statusfile!r})
        '''.format(
            cont=cont, infile=self.infile, outfile=self.resultfile,
            part_outfile=self.partial_resultfile, statusfile=self.statusfile)
        file_dep = [os.path.join(os.path.dirname(self.task.path), f)
                    for f in self.task.file_dep]
        return Job(
            'pspace', self.submit, code, [self.task.path] + file_dep,
            [self.infile])

    # FIXME creates a bunch of identical python files.
    def create_process_job(self):
        jobs = []
        code = '''
from psyrun.backend.load_balancing import LoadBalancingWorker
LoadBalancingWorker({infile!r}, {outfile!r}, {statusfile!r}, task.store).start(
    task.execute)
        '''.format(
            infile=self.infile, outfile=self.partial_resultfile,
            statusfile=self.statusfile)
        for i in range(self.task.max_jobs):
            jobs.append(Job(
                str(i), self.submit, code, [self.infile],
                [self.partial_resultfile]))

        group = JobGroup('process', jobs)
        return group

    def create_finalize_job(self):
        code = '''
import os
os.rename({part!r}, {whole!r})
        '''.format(part=self.partial_resultfile, whole=self.resultfile)
        return Job(
            'finalize', self.submit, code, [self.partial_resultfile],
            [self.resultfile])

    def get_missing(self):
        pspace = self.task.pspace
        try:
            missing_items = missing(
                pspace, Param(**self.task.store.load(self.resultfile)))
        except IOError:
            try:
                missing_items = missing(
                    pspace,
                    Param(**self.task.store.load(self.partial_resultfile)))
            except IOError:
                pass
        return missing_items


class LoadBalancingWorker(object):
    """Maps a function to the parameter space supporting other
    *LoadBalancingWorkers* processing the same input file at the same time.

    Parameters
    ----------
    infile : str
        Filename of the file with the input parameters space.
    outfile : str
        Filename of the file to write the results to.
    statusfile : str
        Filename of the file to track the processing progress.
    store : `Store`, optional
        Input/output backend.
    """
    def __init__(self, infile, outfile, statusfile, store=DefaultStore()):
        self.infile = infile
        self.outfile = outfile
        self.statusfile = statusfile
        self.store = store

    @classmethod
    def create_statusfile(cls, statusfile):
        """Creates the status file required by all load balancing workers.

        Parameters
        ----------
        statusfile : str
            Filename of the status file.
        """
        with open(statusfile, 'w') as f:
            f.write('0')
            f.flush()

    def get_next_ix(self):
        """Get the index of the next parameter assignment to process."""
        with open(self.statusfile, 'r+') as f:
            fcntl.flock(f, fcntl.LOCK_EX)
            try:
                ix = int(f.read())
                f.seek(0)
                f.truncate(0)
                f.write(str(ix + 1))
                f.flush()
            finally:
                fcntl.flock(f, fcntl.LOCK_UN)
        return ix

    def get_next_param_set(self):
        """Load the next parameter assignment to process."""
        return self.store.load(self.infile, row=self.get_next_ix())

    def save_data(self, data):
        """Appends data to the *outfile*.

        Uses a lock on the file to support concurrent access.
        """
        with open(self.statusfile + '.lock', 'w') as lock:
            fcntl.flock(lock, fcntl.LOCK_EX)
            try:
                self.store.append(self.outfile, data)
            finally:
                fcntl.flock(lock, fcntl.LOCK_UN)

    def start(self, fn):
        """Start processing a parameter space.

        A status file needs to be created before invoking this function by
        calling `create_statusfile`.

        Parameters
        ----------
        fn : function
            Function to evaluate on the parameter space.
        """
        while True:
            try:
                pspace = Param(**self.get_next_param_set())
            except IndexError:
                return
            data = map_pspace(fn, pspace)
            self.save_data(data)



"""Base backend interface."""

import os
import sys


class Backend(object):
    """Abstract base class for processing backends.

    Processing backends determine how work is split across jobs.

    Deriving classes are supposed to implement `create_job` and `get_missing`.

    Parameters
    ----------
    task : `TaskDef`
        The task to create processing jobs for.

    Attributes
    ----------
    task : `TaskDef`
        The task to create processing jobs for.
    workdir : str
        Directory in which supporting files for processing the task are stored.
    """

    def __init__(self, task):
        super(Backend, self).__init__()
        self.task = task
        self.workdir = os.path.join(task.workdir, task.name)
        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)

    def submit(self, code, name, depends_on=None):
        """Submits some code to execute to the task scheduler.

        Parameters
        ----------
        code: str
            Code to execute in job.
        name: str
            Job name.
        depends_on: sequence
            Job IDs that have to finish before the submitted code can be
            executed.

        Returns
        -------
        dict
            Contains the id of the submitted job under the key ``'id'``.
        """
        code = '''
try:
    import faulthandler
    faulthandler.enable()
except:
    pass

import os
os.chdir({taskdir!r})

from psyrun.tasks import TaskDef
task = TaskDef({taskpath!r})
{code}
        '''.format(
            path=sys.path,
            taskdir=os.path.abspath(os.path.dirname(self.task.path)),
            taskpath=os.path.abspath(self.task.path), code=code)
        codefile = os.path.join(self.workdir, name + '.py')
        output_filename = os.path.join(self.workdir, name + '.log')
        with open(codefile, 'w') as f:
            f.write(code)

        for job in self.task.scheduler.get_jobs():
            status = self.task.scheduler.get_status(job)
            if status is not None and name == status.name:
                self.task.scheduler.kill(job)

        return self.task.scheduler.submit(
            [self.task.python, codefile], output_filename, name, depends_on,
            self.task.scheduler_args)

    def create_job(self, cont=False):
        """Create the job tree to process given task.

        Parameters
        ----------
        cont : bool, optional
            By default old results will be discarded, but when this option is
            set to True, old results will be kept and merged with the new
            results.
        """
        raise NotImplementedError()

    def get_missing(self):
        """Returns a `ParameterSpace` with missing parameter assignments.

        Missing paramaters assignments are parameter assignments requested by
        the task definition, but that have not been evaluated yet.
        """
        raise NotImplementedError



"""Map functions onto parameter spaces."""

import os.path

from psyrun.pspace import dict_concat, missing, Param


def get_result(fn, params):
    """Evaluates a function with given parameters.

    Evaluates *fn* with the parameters *param* and returns a dictionary with
    the input parameters and returned output values.

    Parameters
    ----------
    fn : function
        Function to evaluate. Has to return a dictionary.
    params : dict
        Parameters passed to *fn* as keyword arguments.

    Returns
    -------
    dict
        Returns *params* updated with the return value of *fn*.

    Examples
    --------
    >>> def fn(x, is_result):
    ...     return {'y': x * x, 'is_result': 1}
    >>>
    >>> from pprint import pprint
    >>> pprint(get_result(fn, {'x': 4, 'is_result': 0}))
    {'is_result': 1, 'x': 4, 'y': 16}
    """
    result = dict(params)
    result.update(fn(**params))
    return result


def map_pspace(fn, pspace):
    """Maps a function to parameter space values.

    Parameters
    ----------
    fn : function
        Function to evaluate on parameter space. Has to return a dictionary.
    pspace : `ParameterSpace`
        Parameter space providing parameter values to evaluate function on.

    Returns
    -------
    dict
        Dictionary with the input parameter values and the function return
        values.

    Examples
    --------
    >>> def fn(x):
    ...     return {'y': x * x}
    >>>
    >>> from pprint import pprint
    >>> from psyrun import Param
    >>> pprint(map_pspace(fn, Param(x=[1, 2])))
    {'x': [1, 2], 'y': [1, 4]}
    """
    return dict_concat(list(get_result(fn, p) for p in pspace.iterate()))


def map_pspace_hdd_backed(fn, pspace, filename, store, return_data=True):
    """Maps a function to parameter space values while storing produced data.

    Data is stored progressively. Thus, if the program crashes, not all data
    will be lost.

    Parameters
    ----------
    fn : function
        Function to evaluate on parameter space. Has to return a dictionary.
    pspace : `ParameterSpace`
        Parameter space providing parameter values to evaluate function on.
    filename : str
        Filename of file to store data to.
    store : `Store`
        Store to save data with.
    return_data : bool, optional
        Whether to return the resulting data after mapping the function. This
        will read all produced data from the disk.

    Returns
    -------
    None or dict
        Dictionary with the input parameter values and the function return
        values if requested.
    """
    if os.path.exists(filename):
        pspace = missing(pspace, Param(**store.load(filename)))
    for p in pspace.iterate():
        store.append(filename, dict_concat((get_result(fn, p),)))
    if not os.path.exists(filename):
        store.save(filename, {})
    if return_data:
        return store.load(filename)


def map_pspace_parallel(fn, pspace, n_jobs=-1, backend='multiprocessing'):
    """Maps a function to parameter space values in parallel.

    Requires `joblib <https://pythonhosted.org/joblib/>`_.

    Parameters
    ----------
    fn : function
        Function to evaluate on parameter space. Has to return a dictionary.
    pspace : ParameterSpace
        Parameter space providing parameter values to evaluate function on.
    n_jobs : int, optional
        Number of parallel jobs. Set to -1 to automatically determine.
    backend : str, optional
        Backend to use. See `joblib documentation
        <https://pythonhosted.org/joblib/parallel.html#using-the-threading-backend>`_
        for details.

    Returns
    -------
    dict
        Dictionary with the input parameter values and the function return
        values.

    Examples
    --------
    >>> from pprint import pprint
    >>> from psyrun import Param
    >>> from psyrun.utils.example import square
    >>>
    >>> pprint(map_pspace_parallel(square, Param(x=[1, 2])))
    {'x': [1, 2], 'y': [1, 4]}
    """
    import joblib
    parallel = joblib.Parallel(n_jobs=n_jobs, backend=backend)
    return dict_concat(parallel(
        joblib.delayed(get_result)(fn, p) for p in pspace.iterate()))



"""NumPy NPZ store."""

import errno

import numpy as np

from psyrun.store.base import Store
from psyrun.store.np_shape import min_shape, match_shape
from psyrun.utils.doc import inherit_docs


@inherit_docs
class NpzStore(Store):
    """Store using NumPy *.npz* files.

    Requires `NumPy <http://www.numpy.org/>`_.

    This store needs to load all data in a file to append to it. Similarly,
    individual rows from a data file can only be returned by loading the whole
    file. Thus, this store is not recommended to be used with applications
    that produce large amounts of data or with the `LoadBalancingBackend` if
    the parameter space is large. It is however more efficient than the
    `PickleStore`.
    """

    ext = '.npz'

    def save(self, filename, data):
        np.savez(filename, **data)

    def load(self, filename, row=None):
        try:
            with np.load(filename) as data:
                if row is None:
                    return dict(data)
                else:
                    return {k: [v[row]] for k, v in data.items()}
        except IOError as err:
            if 'as a pickle' in str(err):
                return {}
            else:
                raise

    def append(self, filename, data):
        try:
            loaded = self.load(filename)
        except IOError as err:
            if err.errno != errno.ENOENT:
                raise
            loaded = {}

        for k, v in data.items():
            shape = min_shape(np.asarray(x).shape for x in v)
            try:
                node = loaded[k]
            except KeyError:
                v = match_shape(v, shape)
                loaded[k] = v
            else:
                shape = min_shape((shape, node.shape[1:]))
                loaded[k] = match_shape(loaded[k], shape)
                v = match_shape(v, shape)
                loaded[k] = np.concatenate((loaded[k], v))
        self.save(filename, loaded)



import os.path

import pytest

from psyrun.store.h5 import H5Store
from psyrun.store.npz import NpzStore
from psyrun.store.pickle import PickleStore
from psyrun.pspace import Param
from psyrun.mapper import (
    map_pspace, map_pspace_parallel, map_pspace_hdd_backed)


def square(a):
    return {'x': a ** 2}


def test_map_pspace():
    calls = []

    def fn(**kwargs):
        calls.append(kwargs)
        return {'result': 42}

    pspace = Param(a=[1, 2])
    result = map_pspace(fn, pspace)

    assert calls == [{'a': 1}, {'a': 2}]
    assert result == {'a': [1, 2], 'result': [42, 42]}


@pytest.mark.parametrize('store', [PickleStore(), H5Store(), NpzStore()])
def test_hdd_backed_mapper(tmpdir, store):
    pspace = Param(a=[1, 2])
    filename = os.path.join(str(tmpdir), 'out' + store.ext)
    result = map_pspace_hdd_backed(
        square, pspace, filename=filename, store=store)
    assert list(result['a']) == [1, 2]
    assert list(result['x']) == [1, 4]
    loaded = store.load(filename)
    assert list(loaded['a']) == [1, 2]
    assert list(loaded['x']) == [1, 4]


@pytest.mark.parametrize('store', [PickleStore(), H5Store(), NpzStore()])
def test_hdd_backed_mapper_continues(tmpdir, store):
    pspace = Param(a=[1, 2])
    filename = os.path.join(str(tmpdir), 'out' + store.ext)
    store.append(filename, {'a': [1], 'x': [-1]})
    result = map_pspace_hdd_backed(
        square, pspace, filename=filename, store=store)
    assert list(result['a']) == [1, 2]
    assert list(result['x']) == [-1, 4]
    loaded = store.load(filename)
    assert list(loaded['a']) == [1, 2]
    assert list(loaded['x']) == [-1, 4]


def test_map_pspace_parallel():
    pspace = Param(a=[1, 2])
    result = map_pspace_parallel(square, pspace)
    assert result == {'a': [1, 2], 'x': [1, 4]}



"""Base store interface."""


class Store(object):
    """Defines the interface of stores.

    Attributes
    ----------
    ext : str
        Filename extension used by the store.
    """

    ext = ''

    def append(self, filename, data):
        """Append data to file.

        When trying to append data to a non-existing file, a new file will be
        created. The backend may require that a file was created with this
        function to be able to append to it.

        Parameters
        ----------
        filename : str
            Filename of file to append the data to.
        data : dict
            Dictionary with data to append.
        """
        raise NotImplementedError()

    def save(self, filename, data):
        """Save data to a file.

        Parameters
        ----------
        filename : str
            Filename of file to save data to.
        data : dict
            Dictionary with data to store.
        """
        raise NotImplementedError()

    def load(self, filename, row=None):
        """Load data from a file.

        Parameters
        ----------
        filename : str
            Filename of file to load data from.
        row : int, optional
            If given, only the row with this index will be loaded.

        Returns
        -------
        dict
            Loaded data.
        """
        raise NotImplementedError()



#!/usr/bin/env python

import imp
import os.path

try:
    from setuptools import find_packages, setup
except ImportError:
    raise ImportError(
        "'setuptools' is required, but not installed. "
        "See https://packaging.python.org/installing/")


version_mod = imp.load_source(
    'version',
    os.path.join(os.path.dirname(__file__), 'psyrun', 'version.py'))

with open('README.rst') as f:
    long_description = f.read()
with open('CHANGES.rst') as f:
    long_description += '\n\n' + f.read()

setup(
    name="psyrun",
    version=version_mod.version,
    author="Jan Gosmann",
    author_email="jan@hyper-world.de",
    url='https://github.com/jgosmann/psyrun',
    license="MIT",
    description="Easy parameter space evaluation and serial farming.",
    long_description=long_description,

    packages=find_packages(),
    provides=['psyrun'],

    install_requires=['six'],
    extras_require={
        'h5': ['numpy', 'tables'],
        'npz': ['numpy'],
        'parallel_map': ['joblib'],
    },

    entry_points={
        'console_scripts': [
            'psy = psyrun.main:psy_main'
        ]
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
)



"""HDF5 store."""

import numpy as np
import tables

from psyrun.store.base import Store
from psyrun.store.np_shape import min_shape, match_shape
from psyrun.utils.doc import inherit_docs


@inherit_docs
class H5Store(Store):
    """Store using the HDF5 format.

    This store allows for efficient appending as long as the dimensions match.

    Requires `NumPy <http://www.numpy.org/>`_ and
    `PyTables <http://www.pytables.org/>`_ . Only numeric data types are
    supported.

    Parameters
    ----------
    node : str, optional
        Node in the HDF5 file to store the data at.
    """

    ext = '.h5'

    def __init__(self, node='/psyrun'):
        super(H5Store, self).__init__()
        self.node = node

    def save(self, filename, data):
        with tables.open_file(filename, 'w') as h5:
            for k, v in data.items():
                self._ensure_supported(v)
                h5.create_array(self.node, k, v, createparents=True)

    def load(self, filename, row=None):
        with tables.open_file(filename, 'r') as h5:
            return {node._v_name: node.read(row) for node in h5.iter_nodes(
                self.node)}

    def append(self, filename, data):
        with tables.open_file(filename, 'a') as h5:
            for k, v in data.items():
                self._ensure_supported(v)
                shape = min_shape(np.asarray(x).shape for x in v)
                try:
                    node = h5.get_node(self.node, k)
                except tables.NoSuchNodeError:
                    v = match_shape(v, shape)
                    h5.create_earray(
                        self.node, k, obj=v, shape=(0,) + shape,
                        createparents=True)
                else:
                    shape = min_shape((shape, node.shape[1:]))
                    v = match_shape(v, shape)
                    if (shape == node.shape[1:] and
                            not isinstance(node.atom, tables.StringAtom)):
                        node.append(v)
                    else:
                        tmp_node = self._get_tmp_node_name(h5)
                        new_node = h5.create_earray(
                            tmp_node, k, atom=tables.Atom.from_dtype(v.dtype),
                            shape=(0,) + shape, createparents=True)
                        for row in node.read():
                            new_node.append(match_shape([row], shape))
                        new_node.append(v)
                        h5.move_node(tmp_node, self.node, k, k, overwrite=True)

    @classmethod
    def _ensure_supported(cls, v):
        dtype = np.asarray(v).dtype
        if dtype.kind in 'OSU':
            raise NotImplementedError(
                "H5Store does not support dtype {}.".format(dtype))

    @staticmethod
    def _get_tmp_node_name(h5):
        i = 0
        while '/tmp{0}'.format(i) in h5:
            i += 1
        return '/tmp{0}'.format(i)



import numpy as np
import pytest

from psyrun.pspace import AmbiguousOperationError, dict_concat, missing, Param


@pytest.mark.parametrize('args,result', [
    (({0: 0}, {0: 1}, {0: 2}), {0: [0, 1, 2]}),
    (({0: 2},), {0: [2]}),
    (({},), {}),
    (({0: 0, 1: 1}, {0: 0, 1: 1}), {0: [0, 0], 1: [1, 1]}),
    ((), {}),
    (({0: 0}, {1: 1}), {0: [0, None], 1: [None, 1]})
])
def test_dict_concat(args, result):
    assert dict_concat(args) == result


class TestParam(object):
    def test_empty(self):
        assert Param().build() == {}

    def test_single(self):
        space = Param(a=[1, 2]).build()
        assert sorted(space['a']) == [1, 2]

    def test_multiple(self):
        space = Param(a=[1, 2], b=[3, 4]).build()
        assert sorted(space['a']) == [1, 2]
        assert sorted(space['b']) == [3, 4]

    def test_multiple_unequal_length_raises_error(self):
        with pytest.raises(ValueError):
            Param(a=[1, 2], b=[3])

    def test_only_scalars(self):
        space = Param(a=1, b=2).build()
        assert space['a'] == [1]
        assert space['b'] == [2]

    def test_treats_string_not_as_sequence(self):
        assert Param(a='item').build()['a'] == ['item']
        assert Param(a=['item']).build()['a'] == ['item']

    def test_scalars_are_broadcasted(self):
        space = Param(a=[1, 2], b=3).build()
        assert sorted(space['a']) == [1, 2]
        assert sorted(space['b']) == [3, 3]

    def test_knows_about_its_keys(self):
        assert sorted(Param(a=[0], b=[1]).keys()) == ['a', 'b']

    def test_length(self):
        assert len(Param(a=[1, 2, 3], b=3)) == 3

    def test_keeps_dtype(self):
        space = Param(a=[1], b=[2.]).build()
        assert np.issubdtype(type(space['a'][0]), int)
        assert np.issubdtype(type(space['b'][0]), float)

    def test_from_dict(self):
        d = {'a': [1, 2], 'b.x': 3, 'b.y': [4, 5]}
        space = Param(**d).build()
        assert space['a'] == [1, 2]
        assert space['b.x'] == [3, 3]
        assert space['b.y'] == [4, 5]

    def test_hierarchical(self):
        d1 = {'root.a': [1, 2], 'b': [3, 4]}
        d2 = {'root.b': [5, 6]}
        space = (Param(**d1) * Param(**d2)).build()
        assert space['root.a'] == [1, 1, 2, 2]
        assert space['b'] == [3, 3, 4, 4]
        assert space['root.b'] == [5, 6, 5, 6]

    def test_str(self):
        p = Param(**{'a': [1, 2], 'b.b': [3, 4]})
        assert str(p) == "Param(a=[1, 2], b.b=[3, 4])"

    def test_rep(self):
        p = Param(**{'a': [1, 2], 'b.b': [3, 4]})
        assert repr(p) == "Param(**{'a': [1, 2], 'b.b': [3, 4]})"


class TestProduct(object):
    def test_product(self):
        space = (Param(a=[1, 2, 3]) * Param(b=[4, 5])).build()
        assert space['a'] == [1, 1, 2, 2, 3, 3]
        assert space['b'] == [4, 5, 4, 5, 4, 5]

    def test_overlapping_product(self):
        with pytest.raises(AmbiguousOperationError):
            _ = Param(a=[1, 2]) * Param(a=[2, 3], b=[4, 5])

    def test_product_with_empty_right(self):
        space = (Param(a=[1, 2, 3]) * Param()).build()
        assert sorted(space['a']) == [1, 2, 3]

    def test_product_with_empty_left(self):
        space = (Param() * Param(a=[1, 2, 3])).build()
        assert sorted(space['a']) == [1, 2, 3]

    def test_product_with_zero_elements(self):
        space = (Param(a=[1, 2, 3]) * Param(b=[])).build()
        assert space['a'] == []
        assert space['b'] == []

    def test_length(self):
        assert len(Param(a=[1, 2, 3]) * Param(b=[1, 2])) == 6
        assert len(Param(a=[1, 2]) * Param()) == 2
        assert len(Param() * Param(a=[1, 2])) == 2
        assert len(Param(a=[1, 2]) * Param(b=[])) == 0
        assert len(Param(a=[]) * Param(b=[1, 2])) == 0

    def test_multidimensional(self):
        space = (Param(a=[[1, 1], [2, 2]]) * Param(b=[1, 2])).build()
        assert space['a'] == [[1, 1], [1, 1], [2, 2], [2, 2]]
        assert space['b'] == [1, 2, 1, 2]


class TestSum(object):
    def test_sum(self):
        space = (Param(a=[1]) + Param(a=[2])).build()
        assert sorted(space['a']) == [1, 2]

    def test_sum_distinct_params(self):
        space = (Param(a=[1]) + Param(b=[2])).build()
        assert space['a'][0] == 1
        assert space['b'][1] == 2
        assert np.isnan(space['a'][1])
        assert np.isnan(space['b'][0])

    def test_sum_with_empty(self):
        space = (Param(a=[1]) + Param()).build()
        assert sorted(space['a']) == [1]

    def test_sum_with_zero_elements(self):
        space = (Param(a=[1]) + Param(b=[])).build()
        assert space['a'] == [1]
        assert np.isnan(space['b'][0])

    def test_length(self):
        assert len(Param(a=[1]) + Param(a=[2, 3])) == 3

    def test_multidimensional(self):
        space = (Param(a=[[1, 1], [2, 2]]) + Param(a=[[3, 3]])).build()
        assert space['a'] == [[1, 1], [2, 2], [3, 3]]

        space = (Param(a=[[1, 1], [2, 2]]) + Param(a=[3, 4])).build()
        assert space['a'] == [[1, 1], [2, 2], 3, 4]


class TestDifference(object):
    def test_difference(self):
        space = (Param(a=[1, 2, 3]) - Param(a=[2])).build()
        assert sorted(space['a']) == [1, 3]

    def test_multiple_builds(self):
        space = (Param(a=[1, 2, 3]) - Param(a=[2]))
        assert sorted(space.build()['a']) == [1, 3]
        assert sorted(space.build()['a']) == [1, 3]

    def test_param_missing_in_subtrahend(self):
        space = (Param(a=[1, 2, 2], b=[4, 5, 6]) - Param(a=[2])).build()
        assert sorted(space['a']) == [1]
        assert sorted(space['b']) == [4]

    def test_param_missign_in_minuend(self):
        with pytest.raises(AmbiguousOperationError):
            _ = Param(a=[1, 2, 3]) - Param(a=[2], b=[3])

    def test_empty_subtrahend(self):
        space = (Param(a=[1, 2, 3]) - Param()).build()
        assert sorted(space['a']) == [1, 2, 3]

    def test_length(self):
        assert len(Param(a=[1, 2, 2], b=[4, 5, 6]) - Param(a=[2])) == 1

    def test_multidimensional(self):
        space = (Param(a=[[1, 1], [2, 2]]) - Param(a=[[2, 2]])).build()
        assert space['a'] == [[1, 1]]


class TestMissing(object):
    def test_missing(self):
        space = missing(Param(a=[1, 2, 3]), Param(a=[2])).build()
        assert sorted(space['a']) == [1, 3]

    def test_param_missing_in_subtrahend(self):
        with pytest.raises(AmbiguousOperationError):
            missing(Param(a=[1, 2, 2], b=[4, 5, 6]), Param(a=[2])).build()

    def test_param_missign_in_minuend(self):
        space = missing(Param(a=[1, 2, 3]), Param(a=[2], b=[3])).build()
        assert sorted(space['a']) == [1, 3]

    def test_empty_subtrahend(self):
        space = missing(Param(a=[1, 2, 3]), Param()).build()
        assert sorted(space['a']) == [1, 2, 3]

    def test_length(self):
        assert len(missing(Param(a=[1, 2, 3]), Param(a=[2]))) == 2

    def test_multidimensional(self):
        space = missing(Param(a=[[1, 1], [2, 2]]), Param(a=[[1, 1]])).build()
        assert space['a'] == [[2, 2]]



"""Backend for distributed parameter evaluation."""

import os

from psyrun.backend.base import Backend
from psyrun.jobs import Job, JobChain, JobGroup
from psyrun.pspace import dict_concat, missing, Param
from psyrun.mapper import map_pspace_hdd_backed
from psyrun.store import DefaultStore
from psyrun.utils.doc import inherit_docs


@inherit_docs
class DistributeBackend(Backend):
    """Create subtasks for distributed parameter evaluation.

    This will create one tasks that splits the parameter space in a number of
    equal batches (at most *max_jobs*, but with at least *min_items* for each
    batch). After processing all batches the results will be merged into a
    single file.

    This is similar to map-reduce processing.

    Parameters
    ----------
    task : `TaskDef`
        Task definition to create subtasks for.
    """

    @property
    def resultfile(self):
        """File in which the results will be stored."""
        if self.task.resultfile:
            return self.task.resultfile
        else:
            return os.path.join(
                self.workdir, 'result' + self.task.store.ext)

    @property
    def pspace_file(self):
        """File that will store the input parameters space."""
        return os.path.join(self.workdir, 'pspace' + self.task.store.ext)

    def _try_mv_to_out(self, filename):
        try:
            os.rename(
                os.path.join(self.workdir, filename),
                os.path.join(self.workdir, 'out', filename))
        except OSError:
            pass

    def create_job(self, cont=False):
        if cont:
            outdir = os.path.join(self.workdir, 'out')
            self._try_mv_to_out('result' + self.task.store.ext)
            Splitter.merge(
                outdir, os.path.join(outdir, 'pre' + self.task.store.ext))
            for filename in os.listdir(outdir):
                if not filename.startswith('pre'):
                    os.remove(os.path.join(outdir, filename))
            pspace = self.get_missing()
        else:
            pspace = self.task.pspace

        self.task.store.save(self.pspace_file, pspace.build())
        splitter = Splitter(
            self.workdir, pspace, self.task.max_jobs, self.task.min_items,
            store=self.task.store)

        split = self.create_split_job(splitter)
        process = self.create_process_job(splitter)
        merge = self.create_merge_job(splitter)
        return JobChain(self.task.name, [split, process, merge])

    def create_split_job(self, splitter):
        code = '''
from psyrun.backend.distribute import Splitter
from psyrun.pspace import Param
pspace = Param(**task.store.load({pspace!r}))
Splitter(
    {workdir!r}, pspace, {max_jobs!r}, {min_items!r},
    store=task.store).split()
        '''.format(
            pspace=self.pspace_file,
            workdir=splitter.workdir, max_jobs=self.task.max_jobs,
            min_items=self.task.min_items)
        file_dep = [os.path.join(os.path.dirname(self.task.path), f)
                    for f in self.task.file_dep]
        return Job(
            'split', self.submit, code, [self.task.path] + file_dep,
            [f for f, _ in splitter.iter_in_out_files()])

    def create_process_job(self, splitter):
        jobs = []
        for i, (infile, outfile) in enumerate(
                splitter.iter_in_out_files()):
            code = '''
from psyrun.backend.distribute import Worker
execute = task.execute
Worker(store=task.store).start(execute, {infile!r}, {outfile!r})
            '''.format(infile=infile, outfile=outfile)
            jobs.append(Job(str(i), self.submit, code, [infile], [outfile]))

        group = JobGroup('process', jobs)
        return group

    def create_merge_job(self, splitter):
        code = '''
from psyrun.backend.distribute import Splitter
Splitter.merge({outdir!r}, {filename!r}, append=False, store=task.store)
        '''.format(outdir=splitter.outdir, filename=self.resultfile)
        return Job(
            'merge', self.submit, code,
            [f for _, f in splitter.iter_in_out_files()], [self.resultfile])

    def get_missing(self):
        pspace = self.task.pspace
        try:
            missing_items = missing(
                pspace, Param(**self.task.store.load(self.resultfile)))
        except IOError:
            missing_items = pspace
            try:
                for filename in os.listdir(os.path.join(self.workdir, 'out')):
                    outfile = os.path.join(self.workdir, 'out', filename)
                    try:
                        x = Param(**self.task.store.load(outfile))
                        missing_items = missing(
                            missing_items,
                            Param(**self.task.store.load(outfile)))
                    except IOError:
                        pass
            except IOError:
                pass
        return missing_items


class Splitter(object):
    """Split a parameter space into multiple input files and merge results
    after processing.

    Parameters
    ----------
    workdir : str
        Working directory to create input files in and read output files from.
    pspace : `ParameterSpace`
        Parameter space to split up.
    max_splits : int, optional
        Maximum number of splits to perform.
    min_items : int, optional
        Minimum number of parameter sets in each split.
    store : `Store`, optional
        Input/output backend.

    Attributes
    ----------
    indir : str
        Directory to store input files.
    max_splits : int
        Maximum number of splits to perform.
    min_items : int
        Minimum number of parameter sets in each split.
    outdir : str
        Directory to store output files.
    pspace : `ParameterSpace`
        Parameter space to split up.
    store : `Store`
        Input/output backend.
    workdir : str
        Working directory to create input files in and read output files from.
    """
    def __init__(
            self, workdir, pspace, max_splits=64, min_items=4,
            store=DefaultStore()):
        self.workdir = workdir
        self.indir = self._get_indir(workdir)
        self.outdir = self._get_outdir(workdir)

        if not os.path.exists(self.indir):
            os.makedirs(self.indir)
        if not os.path.exists(self.outdir):
            os.makedirs(self.outdir)

        self.pspace = pspace
        self.max_splits = max_splits
        self.min_items = min_items

        self.store = store

    @property
    def n_splits(self):
        """Number of total splits that will be generated."""
        n_splits = (len(self.pspace) - 1) // self.min_items + 1
        if self.max_splits is not None:
            n_splits = min(self.max_splits, n_splits)
        return n_splits

    def split(self):
        """Perform splitting of parameters space and save input files for
        processing."""
        items_remaining = len(self.pspace)
        param_iter = self.pspace.iterate()
        for i, filename in enumerate(self._iter_filenames()):
            split_size = self.min_items
            if self.max_splits is not None:
                split_size = max(
                    split_size, items_remaining // (self.max_splits - i))
            items_remaining -= split_size
            block = dict_concat(
                [row for row in self._iter_n(param_iter, split_size)])
            self.store.save(os.path.join(self.indir, filename), block)

    @classmethod
    def merge(cls, outdir, merged_filename, append=True, store=DefaultStore()):
        """Merge processed files together.

        Parameters
        ----------
        outdir : str
            Directory with the output files.
        merged_filename : str
            Filename of file to save with the merged results.
        append : bool, optional
            If True the merged data will be appended, otherwise the file
            will be overwritten with the merged data.
        store : `Store`, optional
            Input/output backend.
        """
        if not append:
            store.save(merged_filename, {})
        for filename in os.listdir(outdir):
            if os.path.splitext(filename)[1] != store.ext:
                continue
            infile = os.path.join(outdir, filename)
            store.append(merged_filename, store.load(infile))

    def iter_in_out_files(self):
        """Return generator returning tuples of corresponding input and output
        filenames."""
        return ((os.path.join(self.indir, f), os.path.join(self.outdir, f))
                for f in self._iter_filenames())

    def _iter_filenames(self):
        return (str(i) + self.store.ext for i in range(self.n_splits))

    @staticmethod
    def _iter_n(it, n):
        for _ in range(n):
            yield next(it)

    @classmethod
    def _get_indir(cls, workdir):
        return os.path.join(workdir, 'in')

    @classmethod
    def _get_outdir(cls, workdir):
        return os.path.join(workdir, 'out')


class Worker(object):
    """Maps a function to the parameter space loaded from a file and writes the
    result to an output file.

    Parameters
    ----------
    store : `Store`, optional
        Input/output backend.

    Attributes
    ----------
    store : `Store`
        Input/output backend.
    """

    def __init__(self, store=DefaultStore()):
        self.store = store

    def start(self, fn, infile, outfile):
        """Start processing a parameter space.

        Parameters
        ----------
        fn : function
            Function to evaluate on the parameter space.
        infile : str
            Parameter space input filename.
        outfile : str
            Output filename for the results.
        """
        pspace = Param(**self.store.load(infile))
        out_root, out_ext = os.path.splitext(outfile)
        map_pspace_hdd_backed(
            fn, pspace, out_root + '.part' + out_ext, store=self.store,
            return_data=False)
        os.rename(out_root + '.part' + out_ext, outfile)
